# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.

import re
from datetime import datetime
from datetime import tzinfo
from datetime import timedelta


class IMLDateTime(datetime):
    @classmethod
    def convert_datetime_to_utc(cls, tm_datetime_local):
        """Convert the local datetime to utc """

        utc_offset_timedelta = cls.now() - cls.utcnow()

        return tm_datetime_local - utc_offset_timedelta

    @classmethod
    def parse(cls, raw_date_str):
        date_match = re.match('.*(\d\d\d\d[-/]\d\d[-/]\d\d)', raw_date_str)
        time_match = re.match('.*(\d\d:\d\d:\d\d)', raw_date_str)
        microsecond_match = re.match('.*\.(\d*)', raw_date_str)
        zone_match = re.match('.*([+-]\d\d:?\d\d)', raw_date_str)

        microseconds = microsecond_match.group(1) if microsecond_match else '0'

        if date_match and time_match:
            naive_date_str = "%s %s.%s" % (date_match.group(1).replace('/', '-'),
                                           time_match.group(1),
                                           microseconds)

            naive_dt = IMLDateTime.strptime(naive_date_str, '%Y-%m-%d %H:%M:%S.%f')
        elif date_match:
            naive_dt = IMLDateTime.strptime(date_match.group(1).replace('/', '-'), '%Y-%m-%d')
        elif time_match:
            naive_time_str = "%s.%s" % (time_match.group(1),
                                        microseconds)

            naive_dt = IMLDateTime.strptime(naive_time_str, '%H:%M:%S.%f')
        else:
            raise ValueError('Unable to parse "%s"' % raw_date_str)

        if zone_match:
            offset_str = zone_match.group(1).replace(':', '')

            offset = int(offset_str[-4:-2]) * 60 + int(offset_str[-2:])

            if offset_str[0] == "-":
                offset = -offset
        else:
            offset = 0

        return naive_dt.replace(tzinfo=FixedOffset(offset))

    @classmethod
    def utcnow(cls):
        return super(IMLDateTime, cls).utcnow().replace(tzinfo=FixedOffset(0))

    @classmethod
    def now(cls, tz=None):
        if tz is None:
            tz = LocalOffset()

        return super(IMLDateTime, cls).now(tz=tz)

    @property
    def as_datetime(self):
        """
        django seems to require an actual datetime - can you believe it rather than a instance of a child. Do they not
        understand inheritance! So return a datetime of the IMLDateTime
        :return: datetime
        """
        return datetime(year=self.year,
                        month=self.month,
                        day=self.day,
                        hour=self.hour,
                        minute=self.minute,
                        second=self.second,
                        microsecond=self.microsecond,
                        tzinfo=self.tzinfo)


class FixedOffset(tzinfo):
    """Fixed offset in minutes: `time = utc_time + utc_offset`."""
    def __init__(self, offset=0):
        self.offset = timedelta(minutes=offset)

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset_value):
        self._offset = offset_value

    def utcoffset(self, dt=None):
        return self._offset

    def tzname(self, dt=None):
        return '<%s>' % self._offset

    def dst(self, dt=None):
        return timedelta(0)

    def __repr__(self):
        minutes = int(self.offset.days * (60 * 24))
        minutes += int(self.offset.seconds / 60)
        minutes += int(self.offset.microseconds / (10 ^ 6)) / 60

        return '%s%02d%02d' % ('-' if minutes < 0 else '',
                               divmod(abs(minutes), 60)[0],
                               divmod(abs(minutes), 60)[1])


class LocalOffset(FixedOffset):
    """Fixed offset to localtime zone. `time = utc_time + local_offset`."""
    def __init__(self):
        utc_offset_timedelta = datetime.now() - datetime.utcnow()

        # because we can't stop time, we will find that the seconds are probably one less and microseconds is not
        # zero. We will have a difference that is something like. 0:59:59.999975
        if utc_offset_timedelta.microseconds != 0:
            utc_offset_timedelta = timedelta(days=utc_offset_timedelta.days,
                                             seconds=utc_offset_timedelta.seconds + 1,
                                             microseconds=0)

        utc_offset_timedelta_minutes = (utc_offset_timedelta.seconds + (utc_offset_timedelta.days * 24 * 3600)) / 60

        super(LocalOffset, self).__init__(utc_offset_timedelta_minutes)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import os
import errno
import time
import itertools
import threading
import platform
from collections import namedtuple
from collections import MutableSequence
import signal


ExpiringValue = namedtuple('ExpiringValue', ['value', 'expiry'])


class ExpiringList(MutableSequence):
    """ Special implementation of a python list which invalidate its elements after a specified 'grace_period' """

    def __init__(self, grace_period):
        self._container = list()
        self.grace_period = grace_period

    def __len__(self):
        return len([x for x in self])

    def __delitem__(self, index):
        del self._container[index]

    def __setitem__(self, index, value):
        self.insert(index, value)

    def __str__(self):
        return str([x for x in self])

    def __getitem__(self, index):
        cur_time = time.time()
        if cur_time > self._container[index].expiry:
            self._container = [x for x in self._container if x.expiry >= cur_time]
        return self._container[index].value

    def insert(self, index, value):
        self._container.insert(index, ExpiringValue(value, time.time() + self.grace_period))


class ExceptionThrowingThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        value = os.environ.get('IML_DISABLE_THREADS', '0')
        # Sometimes not threading helps with debug, disabled by setting IML_DISABLE_THREADS=1
        try:
            self._use_threads = kwargs.pop('use_threads',
                                           True if not value else not bool(int(value)))
        except ValueError:
            self._use_threads = True

        if self._use_threads:
            super(ExceptionThrowingThread, self).__init__(*args, **kwargs)
        self._call_target = kwargs['target']
        self._call_args = kwargs['args']
        self._exception_value = None

    def run(self):
        try:
            return super(ExceptionThrowingThread, self).run()
        except BaseException as e:
            self._exception_value = e

    def start(self):
        if self._use_threads:
            return super(ExceptionThrowingThread, self).start()
        else:
            try:
                self._call_target(*self._call_args)
            except BaseException as e:
                self._exception_value = e

    def join(self):
        if self._use_threads:
            super(ExceptionThrowingThread, self).join()
        if self._exception_value:
            raise self._exception_value

    @classmethod
    def wait_for_threads(cls, threads):
        """Wait for all the threads to finish raising an exception if any of them raise an exception
        We have to capture and then raise one of them because we can't re-raise all of the
        exceptions. We do this to make sure all the threads exit before we start the next test.
        """

        exception_raised = None

        for thread in threads:
            try:
                thread.join()
            except Exception as e:
                exception_raised = e

        if exception_raised:
            raise exception_raised


def all_subclasses(klass):
    """
    :return: All the subclasses of the class passed, scanning the inheritance tree recursively
             to find ALL the subclasses.
    """
    return klass.__subclasses__() + [child for subclass in klass.__subclasses__() for child in
                                     all_subclasses(subclass)]


def enum(*sequential, **named):
    """Return enumerated type object from list of keyword and positional arguments
    Reverse access to Enum identifiers available through 'reverse_mapping' member
    :param sequential: positional arguments
    :param named: keyword arguments
    :return: Enum type object with numbered (sequential) or explicit value (keyword) attributes
    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    reverse = dict((value, key) for key, value in enums.iteritems())
    enums['reverse_mapping'] = reverse
    return type('Enum', (), enums)


PlatformInfo = namedtuple('PlatformInfo', ['system',
                                           'distro',
                                           'distro_version',
                                           'distro_version_full',  # note this returns a string e.g. '6.7.1455'
                                           'python_version_major_minor',
                                           'python_patchlevel',
                                           'kernel_version'])

"""A more readable version of the standard platform commands. Using a named tuple the
usage should be much easier to fathom. Caches the value for speed which presumes
the contents are constant for a given execution.

For a Mac it pretends to be Centos 7.2.

:return: PlatformInfo named tuple
"""
if platform.system() == 'Linux' and platform.linux_distribution()[0] == 'CentOS':
    platform_info = PlatformInfo(platform.system(),
                                 platform.linux_distribution()[0],
                                 float('.'.join(platform.linux_distribution()[1].split('.')[:2])),
                                 platform.linux_distribution()[1],
                                 float("%s.%s" % (platform.python_version_tuple()[0],
                                                  platform.python_version_tuple()[1])),
                                 int(platform.python_version_tuple()[2]),
                                 platform.release())
elif platform.system() == 'Darwin' or platform.system() == 'Linux':
    platform_info = PlatformInfo('Linux',
                                 'CentOS',
                                 '7.2',
                                 '7.2.1551',
                                 float("%s.%s" % (platform.python_version_tuple()[0],
                                                  platform.python_version_tuple()[1])),
                                 int(platform.python_version_tuple()[2]),
                                 '3.10.0-327.36.3.el7.x86_64')
else:
    raise RuntimeError('Unknown system type %s' % platform.system())


class PreserveFileAttributes(object):

    def __init__(self, target):
        self.target = target

    def __enter__(self):
        self.orig_perm = os.stat(self.target).st_mode & 0777
        self.orig_uid = os.stat(self.target).st_uid
        self.orig_gid = os.stat(self.target).st_gid

    def __exit__(self, exception_type, value, _traceback):
        os.chmod(self.target, self.orig_perm)
        os.chown(self.target, self.orig_uid, self.orig_gid)


def wait(timeout, minwait=0.1, maxwait=1.0):
    """
    Generate an exponentially backing-off enumeration with a timeout"

    :param timeout: Number of seconds before iterator times out.
    :param minwait: minimum wait between iterations.
    :param maxwait: maximum wait between iterations.
    :return: No return value
    """
    assert timeout > 0, "Timeout must be > 0"

    for index in itertools.count():
        yield index

        sleep_time = min(minwait, maxwait)
        timeout -= sleep_time

        if timeout < 0:
            break

        time.sleep(sleep_time)
        minwait *= 2


def wait_for_result(lambda_expression, logger, timeout=5 * 60, expected_exception_classes=None):
    """
    Evaluates lambda_expression once/1s until no exceptions matching any of expected_exception_classes or hits
    timeout.

    If timout is reached, exception is re-raised.
    """
    assert timeout > 0, "Timeout must be > 0"

    if expected_exception_classes is None:
        expected_exception_classes = [BaseException]

    running_time = 0

    while True:
        try:
            return lambda_expression()
        except tuple(expected_exception_classes) as e:
            if running_time >= timeout:
                logger.debug("Timed out waiting for command completion assertion was %s" % e)
                raise

        time.sleep(1)
        running_time += 1


def human_to_bytes(value_str):
    """
    Convert something like 1024b, or 1024m to a number of bytes
    Very straight forward takes the index into the conversion strings and uses that as the 1024 power
    """
    conversion = "bkmgtp"

    value = float(value_str[0:-1])
    index = conversion.index(value_str[-1:].lower())

    return int(value * (1024 ** index))


def pid_exists(pid):
    """ Check whether pid exists in the current process table. """
    assert type(pid) is int, 'supplied pid must be a digit'

    if pid < 0:
        return False
    if pid == 0:
        # According to "man 2 kill" PID 0 refers to every process
        # in the process group of the calling process.
        # On certain systems 0 is a valid PID but we have no way
        # to know that in a portable fashion.
        raise ValueError('invalid PID 0')

    try:
        os.kill(pid, signal.SIG_DFL)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True
        else:
            # According to "man 2 kill" possible error values are
            # (EINVAL, EPERM, ESRCH)
            raise
    else:
        return True



#!/usr/bin/env python
# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

excludes = ["*tests*"]

setup(
    name='iml-common',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    license='MIT',
    packages=find_packages(exclude=excludes),
    include_package_data=True,
    author='IML Team',
    url='https://pypi.python.org/pypi/iml-common',
    description='Common library used by multiple IML components',
    long_description=long_description,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='IML lustre high-availability',
)



import mock

from iml_common.lib import util
from iml_common.lib.service_control import ServiceControl
from iml_common.lib.service_control import ServiceControlEL6
from iml_common.lib.service_control import ServiceControlEL7
from iml_common.test.command_capture_testcase import CommandCaptureTestCase
from iml_common.test.command_capture_testcase import CommandCaptureCommand


class TestServiceStateEL6(CommandCaptureTestCase):
    """Test the base class and EL6 subclass functionality, base class only needs to be tested in
    this class, features like retry and callback registration exist in base class
    """

    def setUp(self):
        super(TestServiceStateEL6, self).setUp()

        mock.patch.object(util, 'platform_info', util.PlatformInfo('Linux',
                                                                   'CentOS',
                                                                   0.0,
                                                                   '6.6',
                                                                   0.0,
                                                                   0,
                                                                   '')).start()

        self.test_service = ServiceControl.create('test_service')
        self.assertEqual(type(self.test_service), ServiceControlEL6)

    # Test that the service starts successfully when the start method is called
    def test_service_start_success(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'start')),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status')))

        response = self.test_service.start(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertEqual(self.commands_ran_count, 2)

    # Test that the service stops successfully when the stop method is called
    def test_service_stop_success(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=1, executions_remaining=1))

        response = self.test_service.stop(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertEqual(self.commands_ran_count, 2)

    # Test that the expected outcome is correct when the service starts but keeps failing validation,
    #  when the start method is called
    def test_service_start_validate_fail(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'start')),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=1))

        response = self.test_service.start(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, 'Service test_service is not running after being started')
        self.assertEqual(self.commands_ran_count, 12)

    # Test that the expected outcome is correct when the service fails to start,
    # retries then starts but doesnt pass validation , retries then starts and passes validation
    def test_service_start_initial_fail(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'start'),
                                                rc=1, stderr='Service Failed To Start', executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'start'), executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=1, executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'start'), executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'), executions_remaining=1))
        response = self.test_service.start(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    # Test that the expected outcome is correct when the service fails to stop,
    #  retries then stops but doesnt pass validation , retries then stops and passes validation
    def test_service_stop_initial_fail(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'),
                                                rc=1, stderr='Service Failed To Stop', executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'), executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=0, executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'), executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'), rc=1,
                                                executions_remaining=1))

        response = self.test_service.stop(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    # Test that the expected outcome is correct when the service fails to stop, when the stop method is called
    def test_service_stop_validate_fail(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'stop')),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=0))

        response = self.test_service.stop(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response,
                         'Service test_service is still running after being stopped')
        self.assertEqual(self.commands_ran_count, 12)

    # Test that the expected outcome is correct when the service successful restarts when the restart method is called
    def test_service_restart_success(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'), executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'), rc=1, executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'start'), executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'), executions_remaining=1))

        response = self.test_service.restart(retry_time=0.1, validate_time=0.1)
        self.assertEqual(response, None)
        self.assertRanAllCommandsInOrder()

    # Test that the expected outcome is correct when restart fails by failing to stop the service
    def test_service_restart_failstop(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'),
                                                rc=1, stderr='Service Failed To Stop'))

        response = self.test_service.restart(retry_time=0.1, validate_time=0.1)
        self.assertEqual(response, "Error (1) running '/sbin/service test_service stop': '' 'Service Failed To Stop'")
        self.assertEqual(self.commands_ran_count, 6)

    # Test that the expected outcome is correct when enabling the service
    def test_service_enable_success(self):
        self.add_commands(CommandCaptureCommand(('/sbin/chkconfig', '--add', 'test_service'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/chkconfig', 'test_service', 'on'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/chkconfig', 'test_service'),
                                                executions_remaining=1))

        response = self.test_service.enable()
        self.assertEqual(response, None)

        response = self.test_service.enabled
        self.assertEqual(response, True)
        self.assertRanAllCommandsInOrder()

    # Test that the expected outcome is correct when disabling the service
    def test_service_disable_success(self):
        self.add_commands(CommandCaptureCommand(('/sbin/chkconfig', 'test_service', 'off'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/chkconfig', 'test_service'),
                                                rc=1, executions_remaining=1))

        response = self.test_service.disable()
        self.assertEqual(response, None)

        response = self.test_service.enabled
        self.assertEqual(response, False)
        self.assertRanAllCommandsInOrder()

    def test_service_daemon_reload(self):
        """
        Test the daemon_reload function for ServiceControl.

        By default when using el6 as the test vehicle is does nothing.
        """
        self.test_service.daemon_reload()
        self.assertEqual(self.commands_ran_count, 0)
        self.assertRanAllCommandsInOrder()

    # example callback function
    def example_func(self, service, action_code):
        self.received_codes.append((service,
                                    ServiceControl.ServiceState.reverse_mapping[action_code]))
        return None

    # Test that callback can be registered and that correct messages are received
    def test_service_register_listener_and_receive_success(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'start'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=1, executions_remaining=1))

        self.received_codes = []

        ServiceControl.register_listener('test_service', self.example_func)

        response = self.test_service.start(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertListEqual(self.received_codes, [('test_service', 'SERVICESTARTING'),
                                                   ('test_service', 'SERVICESTARTED')])

        self.received_codes = []
        response = self.test_service.stop(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertListEqual(self.received_codes, [('test_service', 'SERVICESTOPPING'),
                                                   ('test_service', 'SERVICESTOPPED')])

    # Test that callback can be registered and that correct messages are received on control fail
    def test_service_register_listener_and_receive_fail_validation(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'start')),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=1))

        self.received_codes = []

        ServiceControl.register_listener('test_service', self.example_func)

        response = self.test_service.start(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, 'Service test_service is not running after being started')
        self.assertListEqual(self.received_codes, [('test_service', 'SERVICESTARTING'),
                                                   ('test_service', 'SERVICESTARTERROR')])

        # reset so we can test both stop and start fail validation within same test
        self.reset_command_capture()
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'stop')),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'status'),
                                                rc=0))
        self.received_codes = []
        response = self.test_service.stop(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, 'Service test_service is still running after being stopped')
        self.assertListEqual(self.received_codes, [('test_service', 'SERVICESTOPPING'),
                                                   ('test_service', 'SERVICESTOPERROR')])

    # Test that callback can be registered and that correct messages are received on control fail
    def test_service_register_listener_and_receive_fail_initial(self):
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'start'),
                                                rc=1, stderr='Service Failed To Start',
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'start'),
                                                rc=1, stderr='Service Failed To Start',
                                                executions_remaining=1))

        self.received_codes = []

        ServiceControl.register_listener('test_service', self.example_func)

        response = self.test_service.start(retry_time=0.1, validate_time=0.1, retry_count=1)

        self.assertEqual(response, "Error (1) running '/sbin/service test_service start': '' 'Service Failed To Start'")
        self.assertListEqual(self.received_codes, [('test_service', 'SERVICESTARTING'),
                                                   ('test_service', 'SERVICESTARTERROR')])

        # reset so we can test both stop and start fail validation within same test
        self.reset_command_capture()
        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'),
                                                rc=1, stderr='Service Failed To Stop',
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service', 'stop'),
                                                rc=1, stderr='Service Failed To Stop',
                                                executions_remaining=1))

        self.received_codes = []
        response = self.test_service.stop(retry_time=0.1, validate_time=0.1, retry_count=1)

        self.assertEqual(response, "Error (1) running '/sbin/service test_service stop': '' 'Service Failed To Stop'")
        self.assertListEqual(self.received_codes, [('test_service', 'SERVICESTOPPING'),
                                                   ('test_service', 'SERVICESTOPERROR')])

    # Test registering a callback on a service before any ServiceControl instantiated for that
    # service. Also tests unregistering callback function from service notifications.
    def test_register_before_instance(self):
        ServiceControl.register_listener('test_service_2', self.example_func)

        test_service_2 = ServiceControl.create('test_service_2')

        self.add_commands(CommandCaptureCommand(('/sbin/service', 'test_service_2', 'start'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service_2', 'status'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service_2', 'stop'),
                                                executions_remaining=1),
                          CommandCaptureCommand(('/sbin/service', 'test_service_2', 'status'),
                                                rc=1, executions_remaining=1))

        self.received_codes = []

        response = test_service_2.start(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertListEqual(self.received_codes, [('test_service_2', 'SERVICESTARTING'),
                                                   ('test_service_2', 'SERVICESTARTED')])

        self.received_codes = []

        ServiceControl.unregister_listener('test_service_2', self.example_func)

        response = test_service_2.stop(retry_time=0.1, validate_time=0.1)

        self.assertEqual(response, None)
        self.assertListEqual(self.received_codes, [])


class TestServiceStateEL7(CommandCaptureTestCase):
    def setUp(self):
        super(TestServiceStateEL7, self).setUp()

        mock.patch.object(util, 'platform_info', util.PlatformInfo('Linux',
                                                                   'CentOS',
                                                                   0.0,
                                                                   '7.3',
                                                                   0.0,
                                                                   0,
                                                                   '')).start()

        self.test = ServiceControl.create('test_service')
        self.assertEqual(type(self.test), ServiceControlEL7)

    # Test the start method
    def test_service_start(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'start', 'test_service'), executions_remaining=1))

        self.test._start()
        self.assertEqual(('systemctl', 'start', 'test_service'), self._commands_history[0])
        self.assertEqual(self.commands_ran_count, 1)

    # Test the stop method
    def test_service_stop(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'stop', 'test_service'), executions_remaining=1))

        self.test._stop()
        self.assertEqual(('systemctl', 'stop', 'test_service'), self._commands_history[0])
        self.assertEqual(self.commands_ran_count, 1)

    # Test the running method
    def test_service_running(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'is-active', 'test_service'), executions_remaining=1))

        self.test.running
        self.assertEqual(('systemctl', 'is-active', 'test_service'), self._commands_history[0])
        self.assertEqual(self.commands_ran_count, 1)

    # Test the enabled method
    def test_service_enabled(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'is-enabled', 'test_service'), executions_remaining=1))

        self.test.enabled
        self.assertEqual(('systemctl', 'is-enabled', 'test_service'), self._commands_history[0])
        self.assertEqual(self.commands_ran_count, 1)

        # Test the enable method
    def test_service_enable(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'enable', 'test_service'), executions_remaining=1))

        self.test.enable()
        self.assertEqual(('systemctl', 'enable', 'test_service'), self._commands_history[0])
        self.assertEqual(self.commands_ran_count, 1)

    def test_service_daemon_reload(self):
        """
        Test the daemon_reload function for ServiceControl on el7.

        When using el7 it should issue systemctl daemon-reload command.
        """
        self.add_commands(CommandCaptureCommand(('systemctl', 'daemon-reload'), executions_remaining=1))

        self.test.daemon_reload()
        self.assertEqual(self.commands_ran_count, 1)
        self.assertRanAllCommandsInOrder()

        # Test the reload method
    def test_service_reload(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'reload', 'test_service'), executions_remaining=1))

        self.test.reload()
        self.assertEqual(('systemctl', 'reload', 'test_service'), self._commands_history[0])
        self.assertEqual(self.commands_ran_count, 1)

        # Test the disable method
    def test_service_disable(self):
        self.add_commands(CommandCaptureCommand(('systemctl', 'disable', 'test_service'), executions_remaining=1))

        self.test.disable()
        self.assertEqual(('systemctl', 'disable', 'test_service'), self._commands_history[0])
        self.assertEqual(self.commands_ran_count, 1)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.

import os
import errno
import re
import time
import logging

from collections import defaultdict
from lockfile import LockFile, LockTimeout
from ..lib import shell
from blockdevice import BlockDevice
from ..lib.util import pid_exists

try:
    # FIXME: this should be avoided, implicit knowledge of something outside the package
    from chroma_agent.log import daemon_log as log
except ImportError:
    log = logging.getLogger(__name__)

    if not log.handlers:
        handler = logging.FileHandler('blockdevice_zfs.log')
        handler.setFormatter(logging.Formatter("[%(asctime)s: %(levelname)s/%(name)s] %(message)s"))
        log.addHandler(handler)
        log.setLevel(logging.DEBUG)


def get_lockfile_pid(lockfile):
    with open(lockfile, 'r') as f:
        contents = f.readlines()

    assert len(contents) == 1 and contents[0].isdigit(), \
        'unexpected contents of lockfile %s: %s' % (lockfile, contents)

    return int(contents[0])


class ZfsDevice(object):
    """
    This provide two functions.

    Function 1 is to provide locking so that within IML only a single access it made to a zpool device. This in itself
    is not an issue but as code may import/export a device it provides that a zpool is not exported/imported whilst
    something else is using it. It also means we can create atomic functions for zpool changes.

    Function 2 is to provide for code to read the status, attributes etc of a zfs device that is not currently imported
    to the machine. The code imports in read only mode and then exports the device whilst the enclosed code can then
    operate on the device as if it was locally active.
    """

    ZPOOL_LOCK_DIR = '/var/lib/iml/zfs_locks'
    LOCK_ACQUIRE_TIMEOUT = 10

    lock_refcount = defaultdict(int)
    locks_dir_initialized = False

    def __init__(self, device_path, try_import):
        """
        :param device_path: Device path to lock the use of an potential import
        :param try_import: If the device is not imported, import / export it (when used as with). If try_import is false
        then the caller must deal with ensuring the device is accessible.
        """
        self.pool_path = device_path.split(os.path.sep)[0]

        # Scope of a single process
        self.lock = LockFile(os.path.join(self.ZPOOL_LOCK_DIR, self.pool_path))

        # unique identifier used as key for reference counting lock calls on pool/thread/process
        self.lock_unique_id = ':'.join([self.lock.unique_name, self.pool_path])
        self.pool_imported = False
        self.try_import = try_import
        self.available = not try_import         # If we are not going to try to import it, then presume it is available

    def __enter__(self):
        self.lock_pool()

        if self.try_import:
            try:
                imported_pools = shell.Shell.try_run(["zpool", "list", "-H", "-o", "name"]).split('\n')
            except:
                self.unlock_pool()
                raise

            result = None

            if self.pool_path not in imported_pools:
                try:
                    result = self.import_(True)
                    self.pool_imported = (result is None)
                except:
                    self.unlock_pool()
                    raise

            self.available = (result is None)

        return self

    def __exit__(self, type, value, traceback):
        try:
            if self.pool_imported is True:
                result = self.export()

                if result is not None:
                    raise shell.Shell.CommandExecutionError(shell.Shell.RunResult(1, '', result, False),
                                                            ['zpool import of %s' % self.pool_path])

                self.pool_imported = False
        finally:
            self.unlock_pool()

    def lock_pool(self):
        """
        Attempt to acquire a lock on a given zpool through a lockfile, increment the reference count each time
        this method is called, regardless of whether acquire is actually called. Keep trying until we know
        we are locking the zpool.

        Lock acquire polls the timeout at intervals of LOCK_ACQUIRE_TIMEOUT.
        """
        if ZfsDevice.locks_dir_initialized is False:
            # ensure lock directory exists
            try:
                os.makedirs(self.ZPOOL_LOCK_DIR)
            except OSError:
                pass

            ZfsDevice.locks_dir_initialized = True

        while not self.lock.i_am_locking():
            try:
                self.lock.acquire(self.LOCK_ACQUIRE_TIMEOUT)
                with open(self.lock.lock_file, 'w') as f:
                    f.writelines(str(self.lock.pid))
            except LockTimeout:
                # prune locks owned by non-existent processes
                pid = get_lockfile_pid(self.lock.lock_file)

                # validate pid holding lock exists
                if not pid_exists(pid):
                    self.lock.break_lock()
                else:
                    log.warning('lock acquire on lockfile %s timed out, lock owned by PID %s' % (self.lock.lock_file,
                                                                                                 pid))

        self.lock_refcount[self.lock_unique_id] += 1

    def unlock_pool(self):
        """
        Release the held lock, should only be called when holding the lock. Decrement the reference count
        each time this method is called and release when reference count is equal to one.
        """
        self.lock_refcount[self.lock_unique_id] -= 1

        if self.lock_refcount[self.lock_unique_id] == 0:
            self.lock.release()

    def import_(self, readonly):
        """
        This must be called when doing an import as it will lock the device before doing imports and ensure there is
        no confusion about whether a device is import or not.

        :return: None if OK else an error message.
        """
        self.lock_pool()

        try:
            return shell.Shell.run_canned_error_message(['zpool', 'import'] +
                                                        (['-N', '-o', 'readonly=on', '-o', 'cachefile=none'] if readonly else []) +
                                                        [self.pool_path])
        finally:
            self.unlock_pool()

    def export(self):
        """
        This must be called when doing an export as it will lock the device before doing imports and ensure there is
        no confusion about whether a device is import or not.

        Sometimes pools can be busy, in particular it seems when they are imported for short periods of time and then
        exported. Perhaps ZFS is still doing some work on the pool. If the pool is busy retry, every 5 seconds for 1
        minute

        :return: None if OK else an error message.
        """
        self.lock_pool()

        try:
            timeout = (60 / 5)
            result = None

            while timeout > 0:
                result = shell.Shell.run_canned_error_message(['zpool', 'export', self.pool_path])

                if result is None or ('pool is busy' not in result):
                    return result

                timeout -= 1
                time.sleep(5)

            return result
        finally:
            self.unlock_pool()


class NotZpoolException(Exception):
    pass


class BlockDeviceZfs(BlockDevice):
    # From lustre_disk.h
    LDD_F_SV_TYPE_MDT = 0x0001
    LDD_F_SV_TYPE_OST = 0x0002
    LDD_F_SV_TYPE_MGS = 0x0004
    LDD_F_SV_TYPE_MGS_or_MDT = (LDD_F_SV_TYPE_MGS | LDD_F_SV_TYPE_MDT)

    _supported_device_types = ['zfs']

    def __init__(self, device_type, device_path):
        super(BlockDeviceZfs, self).__init__(device_type, device_path)

        self._modules_initialized = False
        self._zfs_properties = None
        self._zpool_properties = None

    def _initialize_modules(self):
        if not self._modules_initialized:
            shell.Shell.try_run(['modprobe', 'osd_zfs'])
            shell.Shell.try_run(['modprobe', 'zfs'])

            self._modules_initialized = True

    def _assert_zpool(self, caller_name):
        if '/' in self._device_path:
            raise NotZpoolException("%s accepts zpools as device_path, '%s' is not a zpool!" % (caller_name,
                                                                                                self._device_path))

    @property
    def filesystem_info(self):
        """
        Verify if any zfs datasets exist on zpool named (self._device_path)

        :return: message describing zfs type and datasets found, None otherwise
        """
        self._assert_zpool('filesystem_info')

        with ZfsDevice(self._device_path, False):
            try:
                device_names = shell.Shell.try_run(['zfs', 'list', '-H', '-o', 'name', '-r', self._device_path]).split('\n')

                datasets = [line.split('/', 1)[1] for line in device_names if '/' in line]

                if datasets:
                    return "Dataset%s '%s' found on zpool '%s'" % ('s' if (len(datasets) > 1) else '',
                                                                   ','.join(datasets),
                                                                   self._device_path)
                return None
            except OSError:                             # zfs not found
                return "Unable to execute commands, check zfs is installed."
            except shell.Shell.CommandExecutionError as e:    # no zpool 'self._device_path' found
                return str(e)

    @property
    def filesystem_type(self):
        """
        Verify if any zfs datasets exist on zpool identified by self._device_path

        :return: 'zfs' if occupied or error encountered, None otherwise
        """
        self._assert_zpool('filesystem_type')

        return self.preferred_fstype if self.filesystem_info is not None else None

    @property
    def uuid(self):
        """
        Try to retrieve the guid property of a zfs device, we will use this as the uuid for block device or file system
        objects.

        :return: uuid of zfs device (usually zpool or dataset)
        """
        out = ""

        try:
            with ZfsDevice(self._device_path, True) as zfs_device:
                if zfs_device.available:
                    out = shell.Shell.try_run(['zfs', 'get', '-H', '-o', 'value', 'guid', self._device_path])
        except OSError:                                     # Zfs not found.
            pass

        lines = [l for l in out.split("\n") if len(l) > 0]

        if len(lines) == 1:
            return lines[0]

        raise RuntimeError("Unable to find UUID for device %s" % self._device_path)

    @property
    def preferred_fstype(self):
        return 'zfs'

    @property
    def failmode(self):
        try:
            self._assert_zpool('failmode')
        except NotZpoolException:
            blockdevice = BlockDevice(self._supported_device_types[0], self._device_path.split('/')[0])

            return blockdevice.failmode
        else:
            with ZfsDevice(self._device_path, False):
                return shell.Shell.try_run(["zpool", "get", "-Hp", "failmode", self._device_path]).split()[2]

    @failmode.setter
    def failmode(self, value):
        try:
            self._assert_zpool('failmode')
        except NotZpoolException:
            blockdevice = BlockDevice(self._supported_device_types[0], self._device_path.split('/')[0])

            blockdevice.failmode = value
        else:
            with ZfsDevice(self._device_path, False):
                shell.Shell.try_run(["zpool", "set", "failmode=%s" % value, self._device_path])

    def zfs_properties(self, reread, log=None):
        """
        Try to retrieve the properties for a zfs device at self._device_path.

        :param reread: Do not use the stored properties, always fetch them from the device
        :param log: optional logger
        :return: dictionary of zfs properties
        """
        if reread or not self._zfs_properties:
            with ZfsDevice(self._device_path, True) as zfs_device:
                self._zfs_properties = {}

                if not zfs_device.available:
                    return self._zfs_properties

                ls = shell.Shell.try_run(["zfs", "get", "-Hp", "-o", "property,value", "all", self._device_path])

                for line in ls.split("\n"):
                    try:
                        key, value = line.split()
                        self._zfs_properties[key] = value
                    except ValueError:                              # Be resilient to things we don't understand.
                        if log:
                            log.info("zfs get for %s returned %s which was not parsable." % (self._device_path, line))
                        pass

        return self._zfs_properties

    def zpool_properties(self, reread, log=None):
        """
        Try to retrieve the properties for a zpool device at self._device_path.

        :param reread: Do not use the stored properties, always fetch them from the device
        :param log: optional logger
        :return: dictionary of zpool properties
        """
        self._assert_zpool('zpool_properties')

        if reread or not self._zpool_properties:
            with ZfsDevice(self._device_path, True) as zfs_device:
                self._zpool_properties = {}

                if not zfs_device.available:
                    return self._zpool_properties

                ls = shell.Shell.try_run(["zpool", "get", "-Hp", "all", self._device_path])

                for line in ls.strip().split("\n"):
                    try:
                        _, key, value, _ = line.split()
                        self._zpool_properties[key] = value
                    except ValueError:                              # Be resilient to things we don't understand.
                        if log:
                            log.info("zpool get for %s returned %s which was not parsable." % (self._device_path, line))
                        pass

        return self._zpool_properties

    def mgs_targets(self, log):
        return {}

    def targets(self, uuid_name_to_target, device, log):
        self._initialize_modules()

        if log:
            log.info("Searching device %s of type %s, uuid %s for a Lustre filesystem" % (device['path'],
                                                                                          device['type'],
                                                                                          device['uuid']))

        zfs_properties = self.zfs_properties(False, log)

        if ('lustre:svname' not in zfs_properties) or ('lustre:flags' not in zfs_properties):
            if log:
                log.info("Device %s did not have a Lustre property values required" % device['path'])
            return self.TargetsInfo([], None)

        # For a Lustre block device, extract name and params
        # ==================================================
        name = zfs_properties['lustre:svname']
        flags = int(zfs_properties['lustre:flags'])

        params = defaultdict(list)

        for zfs_property in zfs_properties:
            if zfs_property.startswith('lustre:'):
                lustre_property = zfs_property.split(':')[1]
                params[lustre_property].extend(re.split(BlockDeviceZfs.lustre_property_delimiters[lustre_property],
                                                        zfs_properties[zfs_property]))

        if name.find("ffff") != -1:
            if log:
                log.info("Device %s reported an unregistered lustre target" % device['path'])
            return self.TargetsInfo([], None)

        if (flags & self.LDD_F_SV_TYPE_MGS_or_MDT) == self.LDD_F_SV_TYPE_MGS_or_MDT:
            # For combined MGS/MDT volumes, synthesise an 'MGS'
            names = ["MGS", name]
        else:
            names = [name]

        return self.TargetsInfo(names, params)

    def import_(self):
        """
        Before importing check the device_path does not reference a dataset, if it does then retry on parent zpool
        block device.

        We can only import the zpool if it's not already imported so check before importing.

        :return: None for success meaning the zpool is imported
        """
        self._initialize_modules()

        try:
            self._assert_zpool('import_')
        except NotZpoolException:
            blockdevice = BlockDevice(self._supported_device_types[0], self._device_path.split('/')[0])

            return blockdevice.import_()

        with ZfsDevice(self._device_path, False) as zfs_device:
            result = zfs_device.import_(False)

            if result is not None and 'a pool with that name already exists' in result:

                if self.zpool_properties(True).get('readonly') == 'on':
                    # Zpool is already imported readonly. Export and re-import writeable.
                    result = self.export()

                    if result is not None:
                        return "zpool was imported readonly, and failed to export: '%s'" % result

                    result = self.import_()

                    if (result is None) and (self.zpool_properties(True).get('readonly') == 'on'):
                        return 'zfs pool %s can only be imported readonly, is it in use?' % self._device_path

                else:
                    # zpool is already imported and writable, nothing to do.
                    return None

            return result

    def export(self):
        """
        Before importing check the device_path does not reference a dataset, if it does then retry on parent zpool
        block device.

        We can only export the zpool if it's already imported so check before exporting.

        :return: None for success meaning the zpool has been exported
        """
        self._initialize_modules()

        try:
            self._assert_zpool('export')
        except NotZpoolException:
            blockdevice = BlockDevice(self._supported_device_types[0], self._device_path.split('/')[0])

            return blockdevice.export()

        with ZfsDevice(self._device_path, False) as zfs_device:
            result = zfs_device.export()

            if result is not None and 'no such pool' in result:
                # Already not imported, nothing to do
                return None

            return result

    @classmethod
    def initialise_driver(cls, managed_mode):
        """
        Enable SPL Multi-Mount Protection for ZFS during failover by generating a hostid to be used by Lustre.

        :return: None on success, error message on failure
        """
        error = None

        # ensure lock directory exists
        try:
            os.makedirs(ZfsDevice.ZPOOL_LOCK_DIR)
        except OSError:
            pass

        if managed_mode is False:
            return error

        if os.path.isfile('/etc/hostid') is False:
            # only create an ID if one doesn't already exist
            result = shell.Shell.run(['genhostid'])

            if result.rc != 0:
                error = 'Error preparing nodes for ZFS multimount protection. gethostid failed with %s' \
                        % result.stderr

        def disable_if_exists(name):
            status = shell.Shell.run(['systemctl', 'status', name])

            if status.rc != 4:
                return shell.Shell.run_canned_error_message([
                    'systemctl',
                    'disable',
                    name
                ])

        if error is None:
            error = reduce(
                lambda x, y: x if x is not None else disable_if_exists(y),
                ['zfs.target',
                 'zfs-import-scan',
                 'zfs-import-cache',
                 'zfs-mount'],
                None
            )

        # https://github.com/zfsonlinux/zfs/issues/3801 describes a case where dkms will not rebuild zfs/spl in the
        # case of an upgrade. The command below ensures that dkms updates zfs/spl after our install which may have lead
        # to a kernel update.
        if error is None:
            for install_package in ['spl', 'zfs']:
                result = shell.Shell.run(['rpm', '-qi', install_package])

                # If we get an error there is no package so nothing to do.
                if result.rc == 0:
                    try:
                        version = next((line.split()[2] for line in result.stdout.split('\n') if line.split()[0] == 'Version'), None)
                    except IndexError:
                            version = None                     # Malformed output so we can't fetch the version.

                    if version is not None:
                        try:
                            error = shell.Shell.run_canned_error_message(['dkms', 'install', '%s/%s' % (install_package, version)])

                            if error is None:
                                error = shell.Shell.run_canned_error_message(['modprobe', install_package])
                        except OSError as e:
                            if e.errno != errno.ENOENT:
                                error = 'Error running "dkms install %s/%s" error return %s' % (install_package, version, e.errno)
                if error:
                    break

        return error

    @classmethod
    def terminate_driver(cls):
        """
        Iterate through existing lockfiles and for each check if pid written into lock is THIS process' pid and
        if so, remove lockfile. If no pid written into lockfile, ignore (linklockfiles).
        """
        lockfile_paths = [os.path.join(ZfsDevice.ZPOOL_LOCK_DIR, name) for name in os.listdir(ZfsDevice.ZPOOL_LOCK_DIR)]

        def validate_or_remove(path):
            try:
                pid = get_lockfile_pid(path)
            except AssertionError:
                pass
            else:
                if os.getpid() == pid:
                    os.remove(path)

        # prune locks owned by THIS process
        map(validate_or_remove, lockfile_paths)

        return None



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import re
import os
import subprocess
from collections import defaultdict
from tempfile import mktemp

from blockdevice import BlockDevice
from ..lib import shell


class BlockDeviceLinux(BlockDevice):
    _supported_device_types = ['linux']
    TARGET_NAME_REGEX = "([\w-]+)-(MDT|OST)\w+"

    def __init__(self, device_type, device_path):
        super(BlockDeviceLinux, self).__init__(device_type, device_path)

        self._modules_initialized = False

    def _initialize_modules(self):
        if not self._modules_initialized:
            try:                                            # osd_ldiskfs will load ldiskfs in Lustre 2.4.0+
                shell.Shell.try_run(['modprobe', 'osd_ldiskfs'])  # TEI-469: Race loading the osd module during mkfs.lustre
            except shell.Shell.CommandExecutionError:
                shell.Shell.try_run(['modprobe', 'ldiskfs'])      # TEI-469: Race loading the ldiskfs module during mkfs.lustre

            self._modules_initialized = True

    @property
    def filesystem_type(self):
        """
        Verify if filesystem exists at self._device_path and return type

        :return: type if exists, None otherwise
        """
        occupying_fs = self._blkid_value("TYPE")

        return occupying_fs

    @property
    def filesystem_info(self):
        """
        Verify if filesystem exists at self._device_path and return message

        :return: message indicating type if exists, None otherwise
        """
        occupying_fs = self._blkid_value("TYPE")

        return None if occupying_fs is None else "Filesystem found: type '%s'" % occupying_fs

    @property
    def uuid(self):
        return self._blkid_value("UUID")

    def _blkid_value(self, value):
        result = shell.Shell.run(["blkid", "-p", "-o", "value", "-s", value, self._device_path])

        if result.rc == 2:
            # blkid returns 2 if there is no filesystem on the device
            return None
        elif result.rc == 0:
            result = result.stdout.strip()

            if result:
                return result
            else:
                # Empty filesystem: blkid returns 0 but prints no FS if it seems something non-filesystem-like
                # like an MBR
                return None
        else:
            raise RuntimeError("Unexpected return code %s from blkid %s: '%s' '%s'" % (result.rc, self._device_path, result.stdout, result.stderr))

    @property
    def preferred_fstype(self):
        return 'ldiskfs'

    def mgs_targets(self, log):
        """
        If there is an MGS in the local targets, use debugfs to get a list of targets.
        Return a dict of filesystem->(list of targets)
        """
        self._initialize_modules()

        result = defaultdict(lambda: [])

        log.info("Searching Lustre logs for filesystems")

        ls = shell.Shell.try_run(["debugfs", "-c", "-R", "ls -l CONFIGS/", self._device_path])
        filesystems = []
        targets = []
        for line in ls.split("\n"):
            try:
                name = line.split()[8]

                match = re.search("([\w-]+)-client", name)
                if match is not None:
                    log.info("Found a filesystem of name %s" % match.group(1).__str__())
                    filesystems.append(match.group(1).__str__())

                match = re.search(self.TARGET_NAME_REGEX, name)
                if match is not None:
                    log.info("Found a target of name %s" % match.group(0).__str__())
                    targets.append(match.group(0).__str__())
            except IndexError:
                pass

        # Read config log "<fsname>-client" for each filesystem
        for fs in filesystems:
            self._read_log("filesystem", fs, "%s-client" % fs, result, log)
            self._read_log("filesystem", fs, "%s-param" % fs, result, log)

        # Read config logs "testfs-MDT0000" etc
        for target in targets:
            self._read_log("target", target, target, result, log)

        return result

    def _read_log(self, conf_param_type, conf_param_name, log_name, result, log):
        # NB: would use NamedTemporaryFile if we didn't support python 2.4
        """
        Uses debugfs to parse information about the filesystem on a device. Return any mgs info
        and config parameters about that device.

        The reality is that nothing makes use of the conf_params anymore but the code existed and this routine
        is unchanged from before 2.0 so I have left conf_params in.

        :param conf_param_type: The type of configuration parameter to store
        :type conf_param_type: str
        :param conf_param_name: The name of the configuration parameter to store
        :type conf_param_name: dict
        :param log_name: The log name to dump the information about dev into
        :type log_name: str
        :param dev: The dev[vice] to parse for log information
        :type dev: str

        Returns: MgsTargetInfo containing targets and conf found.
        """

        tmpfile = mktemp()

        log.info("Reading log for %s:%s from log %s" % (conf_param_type, conf_param_name, log_name))

        try:
            shell.Shell.try_run(["debugfs", "-c", "-R", "dump CONFIGS/%s %s" % (log_name, tmpfile), self._device_path])
            if not os.path.exists(tmpfile) or os.path.getsize(tmpfile) == 0:
                # debugfs returns 0 whether it succeeds or not, find out whether
                # dump worked by looking for output file of some length. (LU-632)
                return

            client_log = subprocess.Popen(["llog_reader", tmpfile], stdout=subprocess.PIPE).stdout.read()

            entries = client_log.split("\n#")[1:]
            for entry in entries:
                tokens = entry.split()
                # ([\w=]+) covers all possible token[0] from
                # lustre/utils/llog_reader.c @ 0f8dca08a4f68cba82c2c822998ecc309d3b7aaf
                (code, action) = re.search("^\\((\d+)\\)([\w=]+)$", tokens[1]).groups()
                if conf_param_type == 'filesystem' and action == 'setup':
                    # e.g. entry="#09 (144)setup     0:flintfs-MDT0000-mdc  1:flintfs-MDT0000_UUID  2:192.168.122.105@tcp"
                    label = re.search("0:([\w-]+)-\w+", tokens[2]).group(1)
                    fs_name = label.rsplit("-", 1)[0]
                    uuid = re.search("1:(.*)", tokens[3]).group(1)
                    nid = re.search("2:(.*)", tokens[4]).group(1)

                    log.info("Found log entry for uuid %s, label %s, nid %s" % (uuid, label, nid))

                    result[fs_name].append({
                        "uuid": uuid,
                        "name": label,
                        "nid": nid})
                elif action == "param" or (action == 'SKIP' and tokens[2] == 'param'):
                    if action == 'SKIP':
                        clear = True
                        tokens = tokens[1:]
                    else:
                        clear = False

                    # e.g. entry="#29 (112)param 0:flintfs-client  1:llite.max_cached_mb=247.9"
                    # has conf_param name "flintfs.llite.max_cached_mb"
                    object = tokens[2][2:]
                    if len(object) == 0:
                        # e.g. "0: 1:sys.at_max=1200" in an OST log: it is a systemwide
                        # setting
                        param_type = conf_param_type
                        param_name = conf_param_name
                    elif re.search(self.TARGET_NAME_REGEX, object):
                        # Identify target params
                        param_type = 'target'
                        param_name = re.search(self.TARGET_NAME_REGEX, object).group(0)
                    else:
                        # Fall through here for things like 0:testfs-llite, 0:testfs-clilov
                        param_type = conf_param_type
                        param_name = conf_param_name

                    if tokens[3][2:].find("=") != -1:
                        key, val = tokens[3][2:].split("=")
                    else:
                        key = tokens[3][2:]
                        val = True

                    if clear:
                        val = None

                    log.info("Found conf param %s:%s:%s of %s" % (param_type, param_name, key, val))

                    # 2.2 don't save the conf params because nothing reads them and zfs doesn't seem to produce them
                    # so keep the code - but just don't store.
                    # This change is being made on the FF date hence the caution.
                    #self.conf_params[param_type][param_name][key] = val
        finally:
            if os.path.exists(tmpfile):
                os.unlink(tmpfile)

    def targets(self, uuid_name_to_target, device, log):
        self._initialize_modules()

        log.info("Searching device %s of type %s, uuid %s for a Lustre filesystem" % (device['path'], device['type'], device['uuid']))

        result = shell.Shell.run(["tunefs.lustre", "--dryrun", device['path']])
        if result.rc != 0:
            log.info("Device %s did not have a Lustre filesystem on it" % device['path'])
            return self.TargetsInfo([], None)

        # For a Lustre block device, extract name and params
        # ==================================================
        name = re.search("Target:\\s+(.*)\n", result.stdout).group(1)
        flags = int(re.search("Flags:\\s+(0x[a-fA-F0-9]+)\n", result.stdout).group(1), 16)
        params_re = re.search("Parameters:\\ ([^\n]+)\n", result.stdout)
        if params_re:
            # Dictionary of parameter name to list of instance values
            params = defaultdict(list)

            # FIXME: naive parse: can these lines be quoted/escaped/have spaces?
            for lustre_property, value in [t.split('=') for t in params_re.group(1).split()]:
                params[lustre_property].extend(re.split(BlockDeviceLinux.lustre_property_delimiters[lustre_property], value))
        else:
            params = {}

        if name.find("ffff") != -1:
            log.info("Device %s reported an unregistered lustre target and so will not be reported" % device['path'])
            return self.TargetsInfo([], None)

        if flags & 0x0005 == 0x0005:
            # For combined MGS/MDT volumes, synthesise an 'MGS'
            names = ["MGS", name]
        else:
            names = [name]

        return self.TargetsInfo(names, params)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import platform
import abc
from collections import namedtuple
from . import util
import re
from ..lib import shell


class FirewallControl(object):
    """Class for issuing shell commands for managing the firewall, this abstract base class is subclassed to
    provide concrete implementations of the abstract methods containing the distribution specific command parameters
    """

    class_override = None
    __metaclass__ = abc.ABCMeta

    platform_use = None

    FirewallRule = namedtuple('FirewallRule', ('port', 'protocol', 'description', 'persist', 'address'))

    # identifiers for results of firewall operations
    SuccessCode = util.enum('UPDATED', 'DUPLICATE', 'NOTRUNNING', 'NORULES')

    def __init__(self, logger=None):
        self.logger = logger
        self.rules = []

    def _log(self, msg, level):
        """ utility function for using reference to an external logger """
        if self.logger and hasattr(self.logger, level):
            getattr(self.logger, level)('{0}: {1}'.format(self.__class__.__name__, msg))

    @classmethod
    def _applicable(cls):
        return cls.platform_use and (util.platform_info.system == 'Linux') and \
               cls.platform_use == util.platform_info.distro_version_full.split('.')[0]

    @classmethod
    def create(cls, logger=None):
        try:
            required_class = next(_cls for _cls in util.all_subclasses(cls) if _cls._applicable())

            return required_class(logger=logger)
        except StopIteration:
            raise RuntimeError('Current platform version not applicable')

    def add_rule(self, port, proto, desc, persist=True, address=None):
        """"Open port(s) in firewall

        :param port: port number
        :param proto: protocol string eg. udp
        :param desc: description of service to be using this port
        :param address: if present, open all ports to this destination address
        :return: error string or None for success
        """
        self._log('Opening firewall for %s' % desc, 'info')
        if address:
            assert port == 0, 'opening a specific port on a source address ' \
                              'is not supported, port value must be 0 (ANY)'

            assert persist is False, 'opening all ports on a source address ' \
                                     'permanently is not currently supported'

            retval = self._add_address(address, proto)
        else:
            assert persist is True, 'opening a single port temporarily is not currently supported'

            retval = self._add_port(port, proto)

        if (retval == self.SuccessCode.UPDATED) and (self.FirewallRule(port, proto, desc, persist, address)
                                                     not in self.rules):
            # only add to list if rule successfully added
            self.rules.append(self.FirewallRule(port, proto, desc, persist, address))

        # if success return code received, return None on success, otherwise return error string
        return None if (retval in self.SuccessCode.reverse_mapping.keys()) else retval

    def remove_rule(self, port, proto, desc, persist=True, address=None):
        """"Close port(s) in firewall
        :param port: port number
        :param proto: protocol string eg. udp
        :param desc: description of service to be using this port
        :param address: if present, remove rule to open all ports to this destination address
        :return: error string or None for success
        """
        self._log('Closing firewall for %s' % desc, 'info')
        if address:
            assert port == 0, 'closing a specific port on a source address is not '\
                              'supported, port value must be 0 (ANY)'

            assert persist is False, 'closing all ports on a source address permanently is not ' \
                                     'currently supported'

            retval = self._remove_address(address, proto)
        else:
            assert persist is True, 'closing a single port temporarily is not currently supported'

            retval = self._remove_port(port, proto)

        if retval == self.SuccessCode.UPDATED:
            # only try to remove from list if rule successfully deleted
            try:
                self.rules.remove(self.FirewallRule(port, proto, desc, persist, address))
            except ValueError:
                pass

        # if success return code received, return None on success, otherwise return error string
        return None if (retval in self.SuccessCode.reverse_mapping.keys()) else retval

    @abc.abstractmethod
    def _add_port(self, port, proto):
        """ Abstract method for opening a single port (to/from all addresses) in firewall """
        raise NotImplementedError

    @abc.abstractmethod
    def _remove_port(self, port, proto):
        """Abstract method for removing rule to open a single port (to/from all addresses) in
        firewall
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _add_address(self, address, proto):
        """ Abstract method for opening all ports on a destination address in firewall """
        raise NotImplementedError

    @abc.abstractmethod
    def _remove_address(self, address, proto):
        """Abstract method for removing rule to open all ports on a destination address in
        firewall
        """
        raise NotImplementedError


class FirewallControlEL6(FirewallControl):
    """ concrete subclass of FirewallControl abstract base class for EL7 firewall control """

    platform_use = '6'

    # return code for 'iptables: Firewall is not running.'
    not_running_rc = 3

    def iptables(self, op, chain, args_in):
        """Verify tables are configured and services running, add supplied arguments to
        rule in specified chain

        :param op: specified operation (add/remove)
        :param chain: specified chain (usually input)
        :param args_in: extra commandline parameters for rule
        :return: error string or None for success
        """
        self._log('Editing iptables...', 'info')

        if op == 'add':
            op_arg = '-I'
        elif op == 'remove':
            op_arg = '-D'
        else:
            raise RuntimeError('invalid mode: %s' % op)

        cmdlist = ''
        rc, stdout, stderr, timeout = shell.Shell.run(['service', 'iptables', 'status'])
        if rc == 0 and stdout != """Table: filter
Chain INPUT (policy ACCEPT)
num  target     prot opt source               destination

Chain FORWARD (policy ACCEPT)
num  target     prot opt source               destination

Chain OUTPUT (policy ACCEPT)
num  target     prot opt source               destination

""":
            # if we are inserting rule into input chain and not specifying an IP address,
            # we want to check if an identical rule exists before adding
            if op_arg == '-I' \
                    and chain == 'INPUT' \
                    and not set(['-s', '--source', '-d', '--destination']).intersection(set(args_in)):

                # parse rule args_in
                state = args_in[args_in.index('--state') + 1].upper()
                proto = args_in[args_in.index('-p') + 1]
                port = args_in[args_in.index('--dport') + 1]
                action = args_in[args_in.index('-j') + 1]

                # separate existing rules
                lines = stdout.split('\n')
                # identify the beginning index of the table, 0 referencing row with table headers
                index = lines.index('Chain INPUT (policy ACCEPT)') + 1

                # additionally match against 'all' protocol identifier in rule spec which would allow traffic of all
                # protocols
                pattern = ' +%s +(%s|all) +.*state %s (%s|all) dpt:%s' % (action, proto, state, proto, port)

                # process all rules within the input chain table until any 'REJECT' rules as we can't reliably
                # assume 'ACCEPT' rules after a 'REJECT' rule will behave as we expect without further analysis
                while lines[index].strip() != '' and lines[index].split()[1] != 'REJECT':

                    if re.search(pattern, lines[index]):
                        # rule already exists at the expected index, don't add again
                        return self.SuccessCode.DUPLICATE

                    index += 1

            cmdlist = ['/sbin/iptables', op_arg, chain]
            cmdlist.extend(args_in)
            rc, stdout, stderr, timeout = shell.Shell.run(cmdlist)

            if rc:
                return stderr or stdout or 'iptables returned unexpected error!'
            else:
                self._log('rule added to iptables: %s' % str(cmdlist), 'debug')
                return self.SuccessCode.UPDATED

        else:
            # indicates firewall un-configured, failed or inactive
            if rc == 0:
                # status returns okay but no rules in chains, un-configured
                self._log('iptables is configured with no rules!', 'warning')
                return self.SuccessCode.NORULES
            elif rc == self.not_running_rc:
                # the IP tables service is not running
                self._log('iptables service is not running!', 'warning')
                return self.SuccessCode.NOTRUNNING
            else:
                return 'service iptables status returned unexpected error!'

    def _add_port(self, port, proto):
        """ EL6 implementation of opening port on firewall using linux shell """
        # XXX using -n and installing the rule manually here is a dirty hack due to lokkit
        # completely restarting the firewall interrupts existing sessions
        error = shell.Shell.run_canned_error_message(['/usr/sbin/lokkit', '-n', '-p', '%s:%s' % (port,
                                                                                           proto)])
        if error:
            return error

        return self.iptables('add', 'INPUT', ['4', '-m', 'state', '--state', 'new', '-p',
                                              proto, '--dport', str(port), '-j', 'ACCEPT'])

    def _remove_port(self, port, proto):
        """ EL6 implementation of closing port on firewall using linux shell """
        # it really bites that lokkit has no "delete" functionality
        error = self.iptables('remove', 'INPUT', ['-m', 'state', '--state', 'new', '-p', proto,
                                                  '--dport', str(port), '-j', 'ACCEPT'])
        import os
        from tempfile import mkstemp
        import shutil
        import errno
        try:
            tmp = mkstemp(dir='/etc/sysconfig')
            with os.fdopen(tmp[0], 'w') as tmpf:
                for line in open('/etc/sysconfig/iptables').readlines():
                    if line.rstrip() != '-A INPUT -m state --state NEW -m %s -p %s --dport %s ' \
                                        '-j ACCEPT' % (proto, proto, port):
                        tmpf.write(line)
                tmpf.flush()
            shutil.move(tmp[1], '/etc/sysconfig/iptables')
        except IOError, e:
            if e.errno != errno.ENOENT:
                raise

        try:
            tmp = mkstemp(dir='/etc/sysconfig')
            with os.fdopen(tmp[0], 'w') as tmpf:
                for line in open('/etc/sysconfig/system-config-firewall').readlines():
                    if line.rstrip() != '--port=%s:udp' % port:
                        tmpf.write(line)
                tmpf.flush()
            shutil.move(tmp[1], '/etc/sysconfig/system-config-firewall')
        except IOError, e:
            if e.errno != errno.ENOENT:
                raise

        return error

    def _add_address(self, daddress, proto):
        """EL6 implementation of opening all ports to a specific destination address on firewall
        using linux shell.Shell. Note that currently this doesn't check for exact duplicates before
        adding
        """
        return self.iptables('add', 'INPUT', ['4', '-m', 'state', '--state', 'NEW', '-m',
                                              proto, '-p', proto, '-d', daddress, '-j', 'ACCEPT'])

    def _remove_address(self, daddress, proto):
        """EL6 implementation of removing rule to open ports to a specific destination address on
        firewall using linux shell
        """
        return self.iptables('remove', 'INPUT', ['-m', 'state', '--state', 'NEW', '-m', proto,
                                                 '-p', proto, '-d', daddress, '-j', 'ACCEPT'])


class FirewallControlEL7(FirewallControl):
    """ concrete subclass of FirewallControl abstract base class for EL7 firewall control """

    platform_use = '7'

    # return code for 'FirewallD is not running'
    not_running_rc = 252
    duplicate_msg = 'Warning: ALREADY_ENABLED\n'

    def _port_rule(self, act, port, proto):
        """Wrapper for port-rule firewall-cmd shell invocations. Command issued twice, once WITH
        the --permanent flag to store persistent settings, and once WITHOUT for immediate effect
        This is the recommended procedure as documented by man page

        Note: rule is activated on the default zone which is assumed is bound to public
        facing interfaces
        """
        arg_list = ['/usr/bin/firewall-cmd', '--%s-port=%s/%s' % (act, port, proto)]

        result = shell.Shell.run(arg_list)

        if result.rc:
            if result.rc == self.not_running_rc:
                # firewall service is not running, this is not an error
                return self.SuccessCode.NOTRUNNING

            return "Error (%s) running '%s': '%s' '%s'" % (result.rc, " ".join(arg_list), result.stdout, result.stderr)

        if result.stdout == self.duplicate_msg:
            return self.SuccessCode.DUPLICATE

        error = shell.Shell.run_canned_error_message(arg_list + ['--permanent'])

        return error or self.SuccessCode.UPDATED

    def _address_rule(self, act, daddress, proto):
        """Wrapper for address-rule firewall-cmd shell invocations. Command issued only for
        immediate effect, in the case of this address rule for EL7 the setting is not persistent
        as white listing all ports on a specific destination address should only be a temporary
        operation so as not to adversely affect security of the target

        Note: rule is activated on the default zone which is assumed to be bound to relevant public
        facing interfaces, rule matches traffic to a destination address on the target
        """
        # temporary firewall rich rule specification
        rich_rule_spec = 'rule family="ipv4" destination address="%s" protocol value="%s" accept' \
                         % (daddress, proto)

        # when constructing the "rich rule" instruction parameter for firewall-cmd we don't have
        # to encapsulate the rule spec in single quotes as you would in bash because shell.Shell.run
        # implicitly assumes it is a value for a single keyword argument
        arg_list = ['/usr/bin/firewall-cmd', '--%s-rich-rule=%s' % (act, rich_rule_spec)]

        result = shell.Shell.run(arg_list)

        if result.rc:
            if result.rc == self.not_running_rc:
                # firewall service is not running, this is not an error
                return self.SuccessCode.NOTRUNNING

            return "Error (%s) running '%s': '%s' '%s'" % (result.rc, " ".join(arg_list), result.stdout, result.stderr)

        if result.stdout == self.duplicate_msg:
            return self.SuccessCode.DUPLICATE

        return self.SuccessCode.UPDATED

    def _add_port(self, port, proto):
        """ EL7 implementation of opening port on firewall using linux shell """
        return self._port_rule('add', port, proto)

    def _remove_port(self, port, proto):
        """ EL7 implementation of removing rule to open port on firewall using linux shell """
        return self._port_rule('remove', port, proto)

    def _add_address(self, address, proto):
        """EL7 implementation of opening all ports to a specific address on firewall using
         linux shell
        """
        return self._address_rule('add', address, proto)

    def _remove_address(self, address, proto):
        """EL7 implementation of removing rule to open ports to a specific address on
        firewall using linux shell
        """
        return self._address_rule('remove', address, proto)


class FirewallControlOSX(FirewallControlEL6):
    """ Just a stub class so that running on OSX things can be made to work.

    We make OSX behave like RH6 because that historically is what happened. So
    this class provides for backwards compatibility.
    """

    # override method to allow source to run on OSX
    @classmethod
    def _applicable(cls):
        return platform.system() == 'Darwin'



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import collections
# OrderedDict is part of the collections module in python2.7
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


NameValueItem = collections.namedtuple("NameValueItem", ['name', 'value'])


# NameValueList provides for list of entities where each one has a name and a value
# they can then be accessed like a dictionary
#
# Create can occur using:
# a dict which contains {'name': name, 'value': value}
# a NameValueItem
# nothing which gives an empty NameValueList
class NameValueList(OrderedDict):
    def __init__(self, entries = []):
        super(NameValueList, self).__init__()

        for entry in entries:
            self.add(entry)

    def add(self, entry):
        if type(entry) == dict:
            if 'name' in entry and 'value' in entry:
                self[entry['name']] = entry['value']
            else:
                for key, value in entry.items():
                    self[key] = value
        elif type(entry) == NameValueItem:
            self[entry.name] = entry.value
        else:
            raise TypeError("%s is not understood" % entry)

    # We have to override keys because it is based on the iterator.
    def keys(self):
        return [item.name for item in self]

    # __iter__ differs because the return is an item with the name and the value
    # See collection below
    def __iter__(self):
        for key in super(NameValueList, self).__iter__():
            yield NameValueItem(key, self[key])

    # Because __iter__ doesn't return just the key we have to overload this as well.
    def iteritems(self):
        for item in self:
            yield (item.name, item.value)

    def collection(self):
        return [{'name': entry.name, 'value': entry.value} for entry in self]



from mock import patch
from iml_common.lib.util import ExpiringList
from iml_common.test.iml_unit_testcase import ImlUnitTestCase


class TestExpiringList(ImlUnitTestCase):

    def setUp(self):
        super(TestExpiringList, self).setUp()
        self.exping_list = ExpiringList(10)
        patcher_time = patch('time.time')
        self.addCleanup(patcher_time.stop)
        self.mock_time = patcher_time.start()
        self.mock_time.return_value = 0

    def test_not_expired(self):
        self.exping_list.append('value')
        self.assertIn('value', self.exping_list)
        self.assertEqual(1, len(self.exping_list))
        self.assertEqual('value', self.exping_list[0])

    def test_expired(self):
        self.exping_list.append('value')
        self.mock_time.return_value = 60 * 10 + 1
        self.assertNotIn('value', self.exping_list)
        self.assertEqual(0, len(self.exping_list))

    def test_deletion(self):
        self.exping_list.append('value 1')
        self.exping_list.append('value 2')
        del self.exping_list[0]
        self.assertEqual(1, len(self.exping_list))
        self.assertEqual('value 2', self.exping_list[0])

    def test_stringify(self):
        self.exping_list.append('value')
        self.assertEqual("['value']", str(self.exping_list))

    def test_multiple_entries(self):
        map(self.exping_list.append, range(1, 101))
        self.assertEqual(100, len(self.exping_list))
        self.mock_time.return_value = 30
        map(self.exping_list.append, range(101, 201))
        self.assertEqual(100, len(self.exping_list))
        self.mock_time.return_value = 60
        self.exping_list.append('value')
        self.assertIn('value', self.exping_list)
        self.assertEqual(1, len(self.exping_list))
        self.assertEqual('value', self.exping_list[0])



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


import ast
import operator

binOps = {
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    ast.Div: operator.div,
    ast.Mod: operator.mod,
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    ast.And: operator.and_,
    ast.Or: operator.or_
}


def safe_eval(expression, properties):
    '''
    safe_eval is an safe expression evaluator. It is fairly well featured and allows for all the operators
    defined abot in binops to be used, it can handle bracketing and understand operator precedence.

    The code takes an expression and a list of properties that will be substituted into the expression.

    For example the expression may be "zfs_installed == False" and the properties might be {"zfs_installed": True}

    Before evaluation this "zfs_installed == False" would be transformed to "1 == 0" - Note that as well as transforming
    the variables True/False and transformed to 1/0.

    The evaluator can return true/false (1/0) or arithmetic answers, it can also deal with string comparisons.

    :param expression: The expression to evaluate.
    :param properties: The properties to be used for the evaluation
    :return: The result of the evaluation or an exception.
    '''

    def substitute_properties(expression, properties):
        for prop, value in properties.items():
            # If the property is a string then it's value needs to be enclosed in quotes.
            if (type(value) == str):
                value = "'%s'" % value

            expression = expression.replace(prop, str(value))

        return expression

    expression = substitute_properties(expression, properties)                  # Sub our properties for their values
    expression = substitute_properties(expression, {'False': 0, 'True': 1})     # Turn False/True into 0/1

    node = ast.parse(expression, mode='eval')

    def _eval(node):
        if isinstance(node, ast.Expression):
            return _eval(node.body)
        elif isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.BinOp):
            return binOps[type(node.op)](_eval(node.left), _eval(node.right))
        elif isinstance(node, ast.BoolOp):
            return binOps[type(node.op)](_eval(node.values[0]), _eval(node.values[1]))
        elif isinstance(node, ast.Compare):
            def _compare(ops, comparators):
                if len(ops) == 1:
                    return binOps[type(ops[0])](_eval(comparators[0]), _eval(comparators[1]))
                return binOps[type(ops[0])](_eval(comparators[0]), _compare(ops[1:], comparators[1:]))
            return _compare(node.ops, [node.left] + node.comparators)
        elif isinstance(node, ast.Name):
            # handle instances of ast.Name, this indicates unrecognised variables in expression
            raise ValueError('Unrecognised variable "{0}"'.format(node.id))
        else:
            raise TypeError('Unsupported type "{0}"'.format(node.__class__))

    return _eval(node.body)



# Copyright (c) 2017 Intel Corporation. All rights reserved.
# Use of this source code is governed by a MIT-style
# license that can be found in the LICENSE file.


from ..lib import shell
from ..lib import util
import abc

_cached_filesystem_types = {}


class FileSystem(object):
    """
    Filesystem abstraction which provides filesystem specific functionality
    This class really really really needs to be in a common place to all code
    so that its functionality can be used in all components. Then we could pass
    it around as a class and not as a hash of its values.
    """

    RC_MOUNT_SUCCESS = 0
    RC_MOUNT_INPUT_OUTPUT_ERROR = 5
    RC_MOUNT_ESHUTDOWN_ERROR = 108

    class_override = None
    __metaclass__ = abc.ABCMeta

    class UnknownFileSystem(KeyError):
        pass

    def __new__(cls, fstype, device_path):
        try:
            # It is possible the caller doesn't know the device type, but they do know the path - these cases should be
            # avoided but in IML today this is the case, so keep a class variable to allow use to resolve it. We default
            # very badly to ldiskfs if we don't have a value.
            if fstype is None:
                if device_path in _cached_filesystem_types:
                    fstype = _cached_filesystem_types[device_path]
                else:
                    fstype = 'ldiskfs'
            else:
                _cached_filesystem_types[device_path] = fstype

            subtype = next(klass for klass in util.all_subclasses(FileSystem) if fstype in klass._supported_filesystems)

            if cls != subtype:
                return subtype.__new__(subtype, fstype, device_path)
            else:
                return object.__new__(cls)

        except StopIteration:
            raise cls.UnknownFileSystem("Filesystem %s unknown" % fstype)

    def __init__(self, fstype, device_path):
        self._fstype = fstype
        self._device_path = device_path

    @abc.abstractmethod
    def label(self):
        """ :return: Returns the label of the filesystem """
        pass

    def device_path(self):
        """ :return: The path to the device the contains/will contain the filesystem """
        return self._device_path

    @property
    def inode_size(self):
        """ :return: The inode size of the filesystem, returns 0 if inode size not supported """
        return 0

    @property
    def inode_count(self):
        """ :return: The inode count of the filesystem, returns 0 if inode count not supported """
        return 0

    def mount(self, mount_point):
        """ :return: Mount the file system, raise an exception on error. """
        self._initialize_modules()

        result = shell.Shell.run(['mount', '-t', 'lustre', self._device_path, mount_point])

        if result.rc == self.RC_MOUNT_INPUT_OUTPUT_ERROR or \
           result.rc == self.RC_MOUNT_ESHUTDOWN_ERROR:
            # HYD-1040, LU-9838: Sometimes we should retry on a failed registration
            result = shell.Shell.run(['mount', '-t', 'lustre', self._device_path, mount_point])

        if result.rc != self.RC_MOUNT_SUCCESS:
            raise RuntimeError("Error (%s) mounting '%s': '%s' '%s'" % (result.rc, mount_point, result.stdout, result.stderr))


    def umount(self):
        """ :return: Umount the file system, raise an exception on error. """
        self._initialize_modules()

        return shell.Shell.try_run(["umount", self._device_path])

    def mount_path(self, target_name):
        """
        :param target_name: The lustre target, ost1, ost2, mgt etc.
        :return: Returns the path to the device for mounting
        """
        return self._device_path

    @abc.abstractmethod
    def mkfs(self, target_name, options):
        """
        Runs mkfs for the filesystem catering for any fs specific irregularities.

        :param target_name: The lustre target, ost1, ost2, mgt etc.
        :param options: mkfs options to be used for the mount.
        :return:
        """
        pass

    @abc.abstractmethod
    def mkfs_options(self, target):
        """
        :param target: The target device that will be formatted.
        :return: A list of filesystem specific options required for the format.
        """
        pass

    @classmethod
    def all_supported_filesystems(self):
        """
        Walks the subclasses to find all the supported filesystems and returns these as a unique list.

        :return: set of supported file systems
        """
        supported_filesystems = []

        for klass in FileSystem.__subclasses__():
            supported_filesystems += klass._supported_filesystems

        return list(set(supported_filesystems))

    @abc.abstractmethod
    def devices_match(self, device1_path, device2_path, device2_uuid):
        """
        Verifies that the devices referenced in the parameters are the same

        :param device1_path: first device string representation
        :param device2_path: second device string representation
        :param device2_uuid: uuid of second device
        :return: return True if both device identifiers reference the same object
        """
        pass
